from builtins import object


class NoneObject(object):
    def __init__(self, description):
        # type: (str) -> None
        self.description = description

    def __getattr__(self, item):
        raise AttributeError("'NoneObject' for '%s' has no attribute '%s'" % (self.description, item))

    def __bool__(self):
        return False



"""The Selene's WebDriver Decorator implementation."""
from abc import ABCMeta, abstractproperty

from future.utils import with_metaclass
from selenium.webdriver.common.by import By
from selenium.webdriver.remote.webdriver import WebDriver

from selene.abctypes.webdriver import IWebDriver
from selene.common.delegation import DelegatingMeta
from selene.common.none_object import NoneObject
from selene.elements import SeleneElement, SeleneCollection


class IWebDriverSource(with_metaclass(ABCMeta, object)):

    @abstractproperty
    def driver(self):
        # type: () -> WebDriver
        pass


# todo: consider implementing it like of DelegatingMeta
class ExplicitWebDriverSource(IWebDriverSource):
    @property
    def driver(self):
        return self._webdriver

    def __init__(self, webdriver):
        self._webdriver = webdriver


class SharedWebDriverSource(IWebDriverSource):

    @property
    def driver(self):
        return self._webdriver

    @driver.setter
    def driver(self, value):
        self._webdriver = value

    def __init__(self):
        self._webdriver = NoneObject("SharedWebDriverSource#_webdriver")  # type: IWebDriver


class SeleneDriver(with_metaclass(DelegatingMeta, IWebDriver)):

    @property
    def __delegate__(self):
        return self._webdriver

    @property
    def _webdriver(self):
        return self._source.driver

    # todo: consider the usage: `SeleneDriver(FirefoxDriver())` over `SeleneDriver.wrap(FirefoxDriver())`
    # todo: it may be possible if __init__ accepts webdriver_or_source and IWebDriverSource implements IWebDriver...
    @classmethod
    def wrap(cls, webdriver):
        # type: (WebDriver) -> SeleneDriver
        return SeleneDriver(ExplicitWebDriverSource(webdriver))

    # def __init__(self, webdriver):
    #     self._webdriver = webdriver

    def __init__(self, webdriver_source):
        # type: (IWebDriverSource) -> None
        self._source = webdriver_source

    def element(self, css_selector_or_by):
        return SeleneElement.by_css_or_by(css_selector_or_by, self)

    s = element
    find = element

    def all(self, css_selector_or_by):
        return SeleneCollection.by_css_or_by(css_selector_or_by, self)

    ss = all
    elements = all
    find_all = all

    # *** SearchContext methods ***
    def find_elements(self, by=By.ID, value=None):
        return self._webdriver.find_elements(by, value)
        # return self.find_all((by, value))

    def find_element(self, by=By.ID, value=None):
        return self._webdriver.find_element(by, value)
        # return self.find((by, value))


_shared_web_driver_source = SharedWebDriverSource()
_shared_driver = SeleneDriver(_shared_web_driver_source)


from abc import ABCMeta, abstractmethod
from _ast import List

from future.utils import with_metaclass
from selenium.webdriver.common.by import By
from selenium.webdriver.remote.webdriver import WebDriver
from selenium.webdriver.remote.webelement import WebElement


class ISearchContext(with_metaclass(ABCMeta, object)):

    @abstractmethod
    def find_element(self, by=By.ID, value=None):
        # type: (By, str) -> WebElement
        raise NotImplementedError

    @abstractmethod
    def find_elements(self, by=By.ID, value=None):
        # type: (By, str) -> List[WebElement]
        raise NotImplementedError

ISearchContext.register(WebDriver)
ISearchContext.register(WebElement)



import warnings
from _ast import Tuple, List
from collections import Sequence

from future.utils import with_metaclass
from selenium.common.exceptions import NoSuchElementException, TimeoutException
from selenium.webdriver import ActionChains
from selenium.webdriver.common.by import By
from selenium.webdriver.common.keys import Keys

import selene
import selene.tools
from selene import config
from selene.abctypes.locators import ISeleneWebElementLocator, ISeleneListWebElementLocator
from selene.abctypes.search_context import ISearchContext
from selene.abctypes.webdriver import IWebDriver
from selene.abctypes.webelement import IWebElement
from selene.common.delegation import DelegatingMeta
from selene.helpers import css_or_by_to_by
from selene.support import by
from selene.support.conditions import be
from selene.support.conditions import have
from selene.wait import wait_for
from selene.conditions import not_, is_matched

try:
    from functools import lru_cache
except ImportError:
    from backports.functools_lru_cache import lru_cache


# todo: consider renaming/refactoring to WebDriverWebElementLocator...
class WebDriverWebElementLocator(ISeleneWebElementLocator):
    def __init__(self, by, search_context):
        # type: (Tuple[By, str], ISearchContext) -> None
        self._by = by
        self._search_context = search_context

    @property
    def description(self):
        return 'first_by%s' % str(self._by)

    def find(self):
        return self._search_context.find_element(*self._by)


class InnerWebElementLocator(ISeleneWebElementLocator):
    def __init__(self, by, element):
        # type: (Tuple[By, str], SeleneElement) -> None
        self._by = by
        self._element = element

    @property
    def description(self):
        return "%s.find_by%s" % (self._element, self._by)

    def find(self):
        # return self._element.get_actual_webelement().find_element(*self._by)
        return wait_for(self._element, be.visible, config.timeout, config.poll_during_waits).find_element(*self._by)


class CachingWebElementLocator(ISeleneWebElementLocator):
    @property
    def description(self):
        return "Caching %s" % (self._element,)

    # todo: will it cash kine of "first wrong webelement"? i.e. invisible element
    @lru_cache()
    def find(self):
        return self._element.get_actual_webelement()

    def __init__(self, element):
        self._element = element


# todo: PyCharm generates abstract methods impl before __init__ method.
# todo: Should we use this order convention? like below...
class IndexedWebElementLocator(ISeleneWebElementLocator):
    def find(self):
        # return self._collection.get_actual_webelements()[self._index]
        return wait_for(
            self._collection,
            have.size_at_least(self._index + 1),
            config.timeout,
            config.poll_during_waits)[self._index]

    @property
    def description(self):
        return "%s[%s]" % (self._collection, self._index)

    def __init__(self, index, collection):
        # type: (int, SeleneCollection) -> None
        self._index = index
        self._collection = collection


class WebdriverListWebElementLocator(ISeleneListWebElementLocator):
    def __init__(self, by, search_context):
        # type: (Tuple[By, str], ISearchContext) -> None
        self._by = by
        self._search_context = search_context

    @property
    def description(self):
        return 'all_by%s' % str(self._by)

    def find(self):
        return self._search_context.find_elements(*self._by)


class InnerListWebElementLocator(ISeleneListWebElementLocator):
    def __init__(self, by, element):
        # type: (Tuple[By, str], SeleneElement) -> None
        self._by = by
        self._element = element

    @property
    def description(self):
        return "(%s).find_all_by(%s)" % (self._element, self._by)

    def find(self):
        # return self._element.get_actual_webelement().find_elements(*self._by)
        return wait_for(self._element, be.visible, config.timeout, config.poll_during_waits)\
            .find_elements(*self._by)


class FilteredListWebElementLocator(ISeleneListWebElementLocator):
    def find(self):
        webelements = self._collection()
        filtered = [webelement
                    for webelement in webelements
                    if is_matched(self._condition, webelement)]
        return filtered

    @property
    def description(self):
        return "(%s).filter_by(%s)" % (self._collection, self._condition.description())

    def __init__(self, condition, collection):
        # type: (Condition, SeleneCollection) -> None
        self._condition = condition
        self._collection = collection


class SlicedListWebElementLocator(ISeleneListWebElementLocator):
    def find(self):
        # webelements = self._collection()
        webelements = wait_for(self._collection,
                               have.size_at_least(self._slice.stop),
                               config.timeout,
                               config.poll_during_waits)
        return webelements[self._slice.start:self._slice.stop:self._slice.step]

    @property
    def description(self):
        return "(%s)[%s:%s:%s]" % (self._collection, self._slice.start, self._slice.stop, self._slice.step)

    def __init__(self, slc,  collection):
        # type: (slice, SeleneCollection) -> None
        self._slice = slc
        self._collection = collection


class FoundByConditionWebElementLocator(ISeleneWebElementLocator):
    def find(self):
        for webelement in self._collection():
            if is_matched(self._condition, webelement):
                return webelement
        raise NoSuchElementException('Element was not found by: %s' % (self._condition,))

    @property
    def description(self):
        return "(%s).select_by(%s)" % (self._collection, self._condition.description())

    def __init__(self, condition, collection):
        # type: (Condition, SeleneCollection) -> None
        self._condition = condition
        self._collection = collection


def _wait_with_screenshot(entity, condition, timeout=None, polling=None):
    if timeout is None:
        timeout = config.timeout
    if polling is None:
        polling = config.poll_during_waits
    try:
        return wait_for(entity, condition, timeout, polling)
    except TimeoutException as e:
        screenshot = selene.tools.take_screenshot()
        msg = '''{original_msg}
            screenshot: {screenshot}'''.format(original_msg=e.msg, screenshot=screenshot)
        raise TimeoutException(msg, e.screen, e.stacktrace)


class SeleneElement(with_metaclass(DelegatingMeta, IWebElement)):

    @property
    def __delegate__(self):
        # type: () -> IWebElement
        return self._locator.find()

    # todo: is this alias needed?
    def get_actual_webelement(self):
        # type: () -> IWebElement
        return self.__delegate__

    # todo: consider removing this method once conditions will be refactored
    # todo: (currently Condition impl depends on this method)
    def __call__(self):
        return self.__delegate__

    # todo: or... maybe better will be remove __delegate__, and just use __call_ instead... ?

    @classmethod
    def by(cls, by, webdriver, context=None):
        # type: (Tuple[str, str], IWebDriver, ISearchContext) -> SeleneElement
        if not context:
            context = webdriver

        return SeleneElement(WebDriverWebElementLocator(by, context), webdriver)

    @classmethod
    def by_css(cls, css_selector, webdriver, context=None):
        # type: (str, IWebDriver, ISearchContext) -> SeleneElement
        if not context:
            context = webdriver

        return SeleneElement.by((By.CSS_SELECTOR, css_selector), webdriver, context)

    @classmethod
    def by_css_or_by(cls, css_selector_or_by, webdriver, context=None):
        if not context:
            context = webdriver

        return SeleneElement.by(
            css_or_by_to_by(css_selector_or_by),
            webdriver,
            context)

    # todo: consider renaming webdriver to driver, because actually SeleneDriver also can be put here...
    def __init__(self, selene_locator, webdriver):
        # type: (ISeleneWebElementLocator, IWebDriver) -> None
        self._locator = selene_locator
        self._webdriver = webdriver
        self._actions_chains = ActionChains(webdriver)

    def __str__(self):
        return self._locator.description

    def _execute_on_webelement(self, command, condition=be.or_not_to_be):
        return command(_wait_with_screenshot(self, condition))

    # *** Relative elements ***

    def element(self, css_selector_or_by):
        return SeleneElement(
            InnerWebElementLocator(css_or_by_to_by(css_selector_or_by), self),
            self._webdriver)

    s = element
    find = element
    # todo: consider making find a separate not-lazy method (not alias)
    # to be used in such example: s("#element").hover().find(".inner").click()
    #                       over: s("#element").hover().element(".inner").click()
    # todo: should then all action-commands return cached elements by default?

    # todo: this is an object, it does not find. should we switch from method to "as a property" implementation?
    def caching(self):
        return SeleneElement(CachingWebElementLocator(self), self._webdriver)

    # todo: cached or cache?
    def cached(self):
        caching = self.caching()
        return caching.should(be.in_dom)

    def all(self, css_selector_or_by):
        # return SeleneCollection.by_css_or_by(css_selector_or_by, self._webdriver, context=self)
        return SeleneCollection(
            InnerListWebElementLocator(css_or_by_to_by(css_selector_or_by), self),
            self._webdriver)

    ss = all
    elements = all
    find_all = all

    @property
    def parent_element(self):
        return self.element(by.be_parent())

    @property
    def following_sibling(self):
        return self.element(by.be_following_sibling())

    @property
    def first_child(self):
        return self.element(by.be_first_child())

    # *** Asserts (Explicit waits) ***

    def should(self, condition, timeout=None):
        if timeout is None:
            timeout = config.timeout
        # todo: implement proper cashing
        # self._found = wait_for(self, condition, timeout)
        _wait_with_screenshot(self, condition, timeout)
        return self

    # todo: consider removing some aliases
    assure = should
    should_be = should
    should_have = should

    def insist(self, condition, timeout=None):
        """ Deprecated use should or assure instead
        """
        warnings.warn("use should or assure instead", DeprecationWarning)
        return self.should(condition, timeout)

    def should_not(self, condition, timeout=None):
        if timeout is None:
            timeout = config.timeout
        # todo: implement proper cashing
        not_condition = not_(condition)
        _wait_with_screenshot(self, not_condition, timeout)
        return self

    # todo: consider removing some aliases
    assure_not = should_not
    should_not_be = should_not
    should_not_have = should_not

    def insist_not(self, condition, timeout=None):
        """ Deprecated use should or assure instead
        """
        warnings.warn("use should_not or assure_not instead", DeprecationWarning)
        return self.should_not(condition, timeout)

    # *** Additional actions ***

    def double_click(self):
        self._execute_on_webelement(
            lambda it: self._actions_chains.double_click(it).perform(),
            condition=be.visible)
        return self

    def set(self, new_text_value):

        def clear_and_send_keys(webelement):
            webelement.clear()
            webelement.send_keys(new_text_value)

        self._execute_on_webelement(
            clear_and_send_keys,
            condition=be.visible)

        return self

    set_value = set

    def press_enter(self):
        return self.send_keys(Keys.ENTER)

    def press_escape(self):
        return self.send_keys(Keys.ESCAPE)

    def press_tab(self):
        return self.send_keys(Keys.TAB)

    def hover(self):
        self._execute_on_webelement(
            lambda it: self._actions_chains.move_to_element(it).perform(),
            condition=be.visible)
        return self

    # *** ISearchContext methods ***

    def find_elements(self, by=By.ID, value=None):
        return self._execute_on_webelement(
            lambda it: it.find_elements(by, value),
            condition=be.visible)
        # return self.__delegate__.find_elements(by, value) # todo: remove

    def find_element(self, by=By.ID, value=None):
        return self._execute_on_webelement(
            lambda it: it.find_element(by, value),
            condition=be.visible)
        # return self.__delegate__.find_element(by, value) # todo: remove

    # *** IWebElement methods ***

    @property
    def tag_name(self):
        return self._execute_on_webelement(
            lambda it: it.tag_name,
            condition=be.in_dom)

    @property
    def text(self):
        return self._execute_on_webelement(
            lambda it: it.text,
            condition=be.visible)

    def click(self):
        self._execute_on_webelement(
            lambda it: it.click(),
            condition=be.visible)
        return self  # todo: think on: IWebElement#click was supposed to return None

    def submit(self):
        self._execute_on_webelement(
            lambda it: it.submit(),
            condition=be.visible)
        return self

    def clear(self):
        self._execute_on_webelement(
            lambda it: it.clear(),
            condition=be.visible)
        return self

    def get_attribute(self, name):
        return self._execute_on_webelement(
            lambda it: it.get_attribute(name),
            condition=be.in_dom)

    def is_selected(self):
        return self._execute_on_webelement(
            lambda it: it.is_selected(),
            condition=be.visible)

    def is_enabled(self):
        return self._execute_on_webelement(
            lambda it: it.is_enabled(),
            condition=be.visible)

    def send_keys(self, *value):
        self._execute_on_webelement(
            lambda it: it.send_keys(*value),
            condition=be.visible)
        return self

    # RenderedWebElement Items
    def is_displayed(self):
        return self._execute_on_webelement(
            lambda it: it.is_displayed(),
            condition=be.in_dom)

    @property
    def location_once_scrolled_into_view(self):
        return self._execute_on_webelement(
            lambda it: it.location_once_scrolled_into_view,
            condition=be.visible)

    @property
    def size(self):
        return self._execute_on_webelement(
            lambda it: it.size,
            condition=be.visible)

    def value_of_css_property(self, property_name):
        return self._execute_on_webelement(
            lambda it: it.value_of_css_property(property_name),
            condition=be.in_dom)

    @property
    def location(self):
        return self._execute_on_webelement(
            lambda it: it.location,
            condition=be.visible)

    @property
    def rect(self):
        return self._execute_on_webelement(
            lambda it: it.rect,
            condition=be.visible)

    @property
    def screenshot_as_base64(self):
        return self._execute_on_webelement(
            lambda it: it.screenshot_as_base64,
            condition=be.visible)  # todo: or `be.in_dom`?

    @property
    def screenshot_as_png(self):
        return self._execute_on_webelement(
            lambda it: it.screenshot_as_png,
            condition=be.visible)  # todo: or `be.in_dom`?

    def screenshot(self, filename):
        return self._execute_on_webelement(
            lambda it: it.screenshot(filename),
            condition=be.visible)  # todo: or `be.in_dom`?

    @property
    def parent(self):
        return self._execute_on_webelement(
            lambda it: it.parent,  # todo: should not we return here some Selene entity as search_context?
            condition=be.in_dom)

    @property
    def id(self):
        return self._execute_on_webelement(
            lambda it: it.id,
            condition=be.in_dom)


class SeleneCollection(with_metaclass(DelegatingMeta, Sequence)):
    """
    To fully match Selenium, SeleneCollection should extend collection.abc.MutableSequence.
    But that's the place where we should be more restrictive.
    It is actually the Selenium, who should use "Sequence" instead of "MutableSequence" (list)
    """

    @property
    def __delegate__(self):
        # type: () -> List[IWebElement]
        return self._locator.find()

    def get_actual_webelements(self):
        # type: () -> List[IWebElement]
        return self.__delegate__

    def __call__(self):
        # type: () -> List[IWebElement]
        return self.__delegate__

    @classmethod
    def by(cls, by, webdriver, context=None):
        # type: (Tuple[str, str], IWebDriver, ISearchContext) -> SeleneCollection
        if not context:
            context = webdriver

        return SeleneCollection(WebdriverListWebElementLocator(by, context), webdriver)

    @classmethod
    def by_css(cls, css_selector, webdriver, context=None):
        # type: (str, IWebDriver, ISearchContext) -> SeleneCollection
        if not context:
            context = webdriver

        return SeleneCollection.by((By.CSS_SELECTOR, css_selector), webdriver, context)

    @classmethod
    def by_css_or_by(cls, css_selector_or_by, webdriver, context=None):
        if not context:
            context = webdriver

        return SeleneCollection.by(css_or_by_to_by(css_selector_or_by), webdriver, context)

    def __init__(self, selene_locator, webdriver):
        # type: (ISeleneListWebElementLocator, IWebDriver) -> None
        self._locator = selene_locator
        self._webdriver = webdriver

    # todo: consider adding self.cashing, self.cashed - like for SeleneElement

    def __str__(self):
        return self._locator.description

    # todo: consider extracting the following not DRY should methods to BaseMixin, or even better: some WaitObject
    # to be mixed in to both Selene Element and Collection
    # Points to think about:
    # * this may break DelegatingMeta logic (because we will have multiple inheritance...)
    # * this will Inheritance... Should not we at least use Composition here?
    def should(self, condition, timeout=None):
        if timeout is None:
            timeout = config.timeout
        _wait_with_screenshot(self, condition, timeout)
        return self

    # todo: consider removing some aliases
    assure = should
    should_be = should
    should_have = should

    def insist(self, condition, timeout=None):
        """ Deprecated use should or assure instead
        """
        warnings.warn("use should or assure instead", DeprecationWarning)
        return self.should(condition, timeout)

    def should_not(self, condition, timeout=None):
        if timeout is None:
            timeout = config.timeout
        # todo: implement proper cashing
        not_condition = not_(condition)
        _wait_with_screenshot(self, not_condition, timeout)
        return self

    # todo: consider removing some aliases are even all of them
    assure_not = should_not
    should_not_be = should_not
    should_not_have = should_not

    def insist_not(self, condition, timeout=None):
        """ Deprecated use should_not or assure_not instead
        """
        warnings.warn("use should_not or assure_not instead", DeprecationWarning)
        return self.should_not(condition, timeout)

    def should_each(self, condition, timeout=None):
        if timeout is None:
            timeout = config.timeout

        for selement in self:
            selement.should(condition, timeout)

    assure_each = should_each

    def should_each_not(self, condition, timeout=None):
        if timeout is None:
            timeout = config.timeout

        for selement in self:
            selement.should_not(condition, timeout)

    assure_each_not = should_each_not

    def filtered_by(self, condition):
        return SeleneCollection(FilteredListWebElementLocator(condition, self), self._webdriver)

    ss = filtered_by
    all_by = filtered_by
    filtered = filtered_by

    filter = filtered_by
    filter_by = filtered_by

    def filterBy(self, condition):
        """ Deprecated use filtered_by or all_by instead
        """
        warnings.warn("use filtered_by or all_by instead", DeprecationWarning)
        return self.filtered_by(condition)

    def element_by(self, condition):
        return SeleneElement(FoundByConditionWebElementLocator(condition, self), self._webdriver)

    s = element_by
    find_by = element_by

    def findBy(self, condition):
        """ Deprecated use element_by instead
        """
        warnings.warn("use element_by instead", DeprecationWarning)
        return self.element_by(condition)

    def find(self, condition):
        """ Deprecated use element_by instead
        """
        warnings.warn("use element_by instead", DeprecationWarning)
        return self.element_by(condition)

    # *** Sequence methods ***

    def __getitem__(self, index):
        if isinstance(index, slice):
            return SeleneCollection(
                SlicedListWebElementLocator(index, collection=self),
                self._webdriver)
        return SeleneElement(IndexedWebElementLocator(index, collection=self), self._webdriver)

    def __len__(self):
        # todo: optimise to the following:
        #   return self.waifFor(size_at_least(0)),
        # where waitFor will return the result of condition application, not self like should
        return len(_wait_with_screenshot(self, have.size_at_least(0)))

    # *** Overriden Sequence methods ***

    def __iter__(self):
        i = 0
        current_len = len(self)
        while i < current_len:
            v = self[i]
            yield v
            i += 1

    # *** Additional Collection style methods ***

    # *** Useful shortcuts ***

    def size(self):
        return len(self)

    def first(self):
        return self[0]



from selene import conditions

size = conditions.size
size_at_least = conditions.size_at_least
text = conditions.text
exact_text = conditions.exact_text
texts = conditions.texts
exact_texts = conditions.exact_texts
css_class = conditions.css_class
attribute = conditions.attribute
value = conditions.value
not_ = conditions.not_


def js_returned_true(script_to_return_bool):
    return conditions.JsReturnedTrue(script_to_return_bool)


def title(text):
    return conditions.Title(text)



import os

from selenium.webdriver.remote.webdriver import WebDriver

import selene.driver
import selene.factory
import selene.config
from selene.elements import SeleneElement, SeleneCollection
from selene.wait import wait_for


def quit_driver():
    get_driver().quit()


def set_driver(driver):
    # type: (WebDriver) -> None
    if selene.factory.is_another_driver(driver):
        selene.factory.kill_all_started_drivers()
    selene.factory.set_shared_driver(driver)


def get_driver():
    # type: () -> WebDriver
    return selene.factory.ensure_driver_started(selene.config.browser_name)


def visit(absolute_or_relative_url):
    """
    Loads a web page in the current browser session.
    :param absolute_or_relative_url:
        an absolute url to web page in case of config.app_host is not specified,
        otherwise - relative url correspondingly

    :Usage:
        visit('http://mydomain.com/subpage1')
        visit('http://mydomain.com/subpage2')
        # OR
        config.app_host = 'http://mydomain.com'
        visit('/subpage1')
        visit('/subpage2')
    """
    get_driver().get(selene.config.app_host + absolute_or_relative_url)


def s(css_selector_or_by):
    # return SElement(css_selector_or_locator)
    return SeleneElement.by_css_or_by(css_selector_or_by, selene.driver._shared_driver)


element = s


def ss(css_selector_or_by):
    # return SElementsCollection(css_selector_or_locator, of=of)
    return SeleneCollection.by_css_or_by(css_selector_or_by, selene.driver._shared_driver)


elements = ss


def take_screenshot(path=None, filename=None):
    if not path:
        path = selene.config.screenshot_folder
    if not filename:
        filename = "screen_{id}".format(id=next(selene.config.counter))
    screenshot_path = os.path.join(path,
                                   "{}.png".format(filename))

    folder = os.path.dirname(screenshot_path)
    if not os.path.exists(folder):
        os.makedirs(folder)

    get_driver().get_screenshot_as_file(screenshot_path)
    return screenshot_path


# todo: consider adding aliases, like: wait_until, wait_brhwser_to
def wait_to(webdriver_condition, timeout=None, polling=None):
    if timeout is None:
        timeout = selene.config.timeout
    if polling is None:
        polling = selene.config.poll_during_waits

    return wait_for(get_driver(), webdriver_condition, timeout, polling)


def execute_script(script, *args):
    return get_driver().execute_script(script, *args)



class ConditionMismatchException(Exception):
    """
    """

    def __init__(self, message='condition did not match', expected=None, actual=None):
        super(ConditionMismatchException, self).__init__()
        self.message = message
        if expected is not None:
            self.message += '''
            \texpected: {}'''.format(expected)
        if actual is not None:
            self.message += '''
            \t  actual: {}'''.format(actual)



import atexit

from selenium import webdriver
from webdriver_manager.chrome import ChromeDriverManager
from webdriver_manager.firefox import GeckoDriverManager
from selenium.webdriver.remote.webdriver import WebDriver

import selene
import selene.driver
from selene import config
from selene.browsers import Browser
import selene.tools


def set_shared_driver(driver):
    selene.driver._shared_web_driver_source.driver = driver
    config.browser_name = driver.name


def get_shared_driver():
    return selene.driver._shared_web_driver_source.driver


def is_another_driver(driver):
    try:
        return get_shared_driver().session_id != driver.session_id
    except AttributeError:
        return False


def is_driver_still_open(driver):
    # type: (WebDriver) -> bool
    try:
        driver.title
    # todo: specify exception?.. (unfortunately there Selenium does not use some specific exception for this...)
    except Exception:
        return False
    return True


def driver_has_started(name):
    shared_driver = get_shared_driver()
    if not shared_driver:
        return False
    return shared_driver.name == name \
        and shared_driver.session_id \
        and is_driver_still_open(shared_driver)


def kill_all_started_drivers():
    atexit._run_exitfuncs()


def ensure_driver_started(name):
    if driver_has_started(name):
        return get_shared_driver()

    return _start_driver(name)


def _start_driver(name):
    kill_all_started_drivers()
    if name == Browser.CHROME:
        driver = webdriver.Chrome(executable_path=ChromeDriverManager().install())
    elif name == Browser.MARIONETTE:
        driver = webdriver.Firefox(executable_path=GeckoDriverManager().install())
    else:
        driver = webdriver.Firefox()
    set_shared_driver(driver)
    _register_driver(driver)
    return driver


def _register_driver(driver):
    atexit.register(driver.quit)



from selene import conditions

visible = conditions.visible
in_dom = conditions.in_dom
enabled = conditions.enabled
empty = conditions.empty
existing = conditions.exist
clickable = conditions.clickable
hidden = conditions.hidden
blank = conditions.blank
or_not_to_be = conditions.or_not_to_be
not_ = conditions.not_



import ast
from _ast import List
from abc import ABCMeta, abstractmethod
import operator

from future.utils import with_metaclass, lmap

from selene.abctypes.conditions import IEntityCondition
from selene.abctypes.webdriver import IWebDriver
from selene.abctypes.webelement import IWebElement
# from selene.elements import SeleneElement, SeleneCollection
from selene.exceptions import ConditionMismatchException


class OrNotToBe(IEntityCondition):

    def description(self):
        return self.__class__.__name__

    def fn(self, entity):
        return entity


or_not_to_be = OrNotToBe


class Not(IEntityCondition):

    def __init__(self, condition):
        # type: (IEntityCondition) -> None
        self._condition = condition

    def description(self):
        return 'not {}'.format(self._condition.description())

    def fn(self, entity):
        try:
            self._condition.fn(entity)
        except Exception as reason:
            return reason
        raise ConditionMismatchException()  # todo: add more information to message


not_ = Not


# *** WebDriver Conditions ***

class WebDriverCondition(with_metaclass(ABCMeta, IEntityCondition)):
    @abstractmethod
    def fn(self, webdriver):
        pass

    def description(self):
        return self.__class__.__name__


class JsReturnedTrue(WebDriverCondition):

    def __init__(self, script_to_return_bool):
        self.script = script_to_return_bool

    def fn(self, webdriver):
        # type: (IWebDriver) -> bool
        result = webdriver.execute_script(self.script)
        if not result:
            raise ConditionMismatchException(
                expected='''script: {script}
                \t\t to return: true'''.format(script=self.script),
                actual='''returned: {result}'''.format(result=result))

js_returned_true = JsReturnedTrue


class Title(WebDriverCondition):

    def __init__(self, text):
        self.expected = text

    def fn(self, webdriver):
        # type: (IWebDriver) -> bool
        actual = webdriver.title
        if not self.expected == actual:
            raise ConditionMismatchException(
                expected=self.expected,
                actual=actual)


title = Title

# *** Element Conditions ***


class ElementCondition(with_metaclass(ABCMeta, IEntityCondition)):

    def description(self):
        return self.__class__.__name__

    def fn(self, element):
        # type: (SeleneElement) -> IWebElement
        return self.match(element.get_actual_webelement())

    @abstractmethod
    def match(self, webelement):
        # type: (IWebElement) -> IWebElement
        pass


def is_matched(condition, webelement):
    # type: (ElementCondition, IWebElement) -> bool
    try:
        condition.match(webelement)
        return True
    except Exception:
        return False


class Visible(ElementCondition):
    def match(self, webelement):
        # type: (SeleneElement) -> IWebElement
        if not webelement.is_displayed():
            raise ConditionMismatchException()
        return webelement


visible = Visible()
appear = visible


class Hidden(ElementCondition):
    def match(self, webelement):
        # type: (SeleneElement) -> IWebElement
        if webelement.is_displayed():
            raise ConditionMismatchException()
        return webelement


hidden = Hidden()
disappear = hidden


# todo: consider removing this condition... because it can confuse somebody...
# it's actually kind of "pseudo-clickable", the actual "clackability" depends on js events...
# todo: implement as and_(displayed, enabled)
class Clickable(ElementCondition):
    def match(self, webelement):
        # type: (IWebElement) -> IWebElement
        actual_displayed = webelement.is_displayed()
        actual_enabled = webelement.is_enabled()
        if not (actual_displayed and actual_enabled):
            raise ConditionMismatchException(
                expected='displayed and enabled',
                actual='displayed: {displayed}, enabled: {enabled}'.format(
                    displayed=actual_displayed, enabled=actual_enabled))
        return webelement


clickable = Clickable()


class Enabled(ElementCondition):
    def match(self, webelement):
        # type: (SeleneElement) -> IWebElement
        if not webelement.is_enabled():
            raise ConditionMismatchException()
        return webelement


enabled = Enabled()


class InDom(ElementCondition):
    """
    checks if element exist in DOM
    """
    def match(self, webelement):
        return webelement


in_dom = InDom()
exist = in_dom


class Text(ElementCondition):

    def __init__(self, expected_text):
        self.expected_text = expected_text

    def match(self, webelement):
        actual_text = webelement.text
        if self.expected_text not in actual_text:
            raise ConditionMismatchException(expected=self.expected_text, actual=actual_text)
        return webelement


text = Text


class ExactText(ElementCondition):

    def __init__(self, expected_text):
        self.expected_text = expected_text

    def match(self, webelement):
        actual_text = webelement.text
        if not self.expected_text == actual_text:
            raise ConditionMismatchException(expected=self.expected_text, actual=actual_text)
        return webelement

exact_text = ExactText


class CssClass(ElementCondition):

    def __init__(self, expected):
        self.expected = expected

    def match(self, webelement):
        actual = webelement.get_attribute("class")
        if self.expected not in actual.split():
            raise ConditionMismatchException(expected=self.expected, actual='class attribute: {}'.format(actual))
        return webelement


css_class = CssClass


class Attribute(ElementCondition):

    def __init__(self, name, value):
        self.name = name
        self.value = value

    def match(self, webelement):
        actual = webelement.get_attribute(self.name)
        if not self.value == actual:
            raise ConditionMismatchException(
                expected='{name}="{value}"'.format(name=self.name, value=self.value),
                actual='{name}="{value}"'.format(name=self.name, value=actual))
        return webelement

attribute = Attribute


def value(val):
    return Attribute('value', val)


blank = value('')


# *** Collection Conditions ***


class CollectionCondition(with_metaclass(ABCMeta, IEntityCondition)):

    def description(self):
        return self.__class__.__name__

    def fn(self, elements):
        # type: (SeleneCollection) -> List[IWebElement]
        return self.match(elements.get_actual_webelements())

    @abstractmethod
    def match(self, webelements):
        # type: (List[IWebElement]) -> List[IWebElement]
        pass


class Texts(CollectionCondition):

    def __init__(self, *expected):
        self.expected = expected

    def match(self, webelements):
        actual = [it.text for it in webelements]
        if not (len(actual) == len(self.expected) and all(lmap(operator.contains, actual, self.expected))):
            raise ConditionMismatchException(
                expected=self.expected,
                actual=actual)
        return webelements

texts = Texts


class ExactTexts(CollectionCondition):

    def __init__(self, *expected):
        self.expected = expected

    def match(self, webelements):
        actual = [it.text for it in webelements]
        if not (len(actual) == len(self.expected) and all(lmap(operator.eq, actual, self.expected))):
            raise ConditionMismatchException(
                expected=self.expected,
                actual=actual)
        return webelements

exact_texts = ExactTexts


class Size(CollectionCondition):
    def __init__(self, expected):
        self.expected = expected

    def match(self, webelements):
        actual = len(webelements)
        if not actual == self.expected:
            raise ConditionMismatchException(
                expected=self.expected,
                actual=actual)
        return webelements

size = Size
empty = size(0)


class SizeAtLeast(CollectionCondition):
    def __init__(self, expected):
        self.expected = expected

    def match(self, webelements):
        actual = len(webelements)
        if not actual >= self.expected:
            raise ConditionMismatchException(
                expected='>= {}'.format(self.expected),
                actual=actual)
        return webelements

size_at_least = SizeAtLeast



from abc import ABCMeta, abstractmethod

from future.utils import with_metaclass


class IEntityCondition(with_metaclass(ABCMeta, object)):

    # todo: consider using __call__ instead
    @abstractmethod
    def fn(self, entity):
        # type: (object) -> any
        pass

    def __call__(self, entity):
        return self.fn(entity)

    @abstractmethod
    def description(self):
        # type: () -> str
        pass


# todo: make the properties also "object oriented" to support different configs per different SeleneDriver instances
import itertools
import os
import time

from selene.browsers import Browser

timeout = 4
poll_during_waits = 0.1
app_host = ''

# todo: make cashing work (currently will not work...)
cash_elements = False
"""To cash all elements after first successful find
      config.cash_elements = True"""

browser_name = Browser.FIREFOX

counter = itertools.count(start=int(round(time.time() * 1000)))

screenshot_folder = os.path.join(os.path.expanduser("~"),
                                 ".selene",
                                 "screenshots",
                                 str(next(counter)))



import selene
import selene.tools
from selene import config
# from selene.elements import SeleneElement, SeleneCollection

import time
from selenium.common.exceptions import TimeoutException

from selene.abctypes.conditions import IEntityCondition
from selene.exceptions import ConditionMismatchException


def wait_for(entity, condition, timeout=4, polling=0.1):
    # type: (object, IEntityCondition, int) -> object
    end_time = time.time() + timeout
    while True:
        try:
            return condition.fn(entity)
        except Exception as reason:
            reason_message = getattr(reason, 'msg',
                                     getattr(reason, 'message',
                                             getattr(reason, 'args', '')))
            reason_string = '{name}: {message}'.format(name=reason.__class__.__name__, message=reason_message)
            screen = getattr(reason, 'screen', None)
            stacktrace = getattr(reason, 'stacktrace', None)

            if time.time() > end_time:
                raise TimeoutException('''
            failed while waiting {timeout} seconds
            to assert {condition}
            for {entity}

            reason: {reason}'''.format(
                    timeout=timeout,
                    condition=condition.description(),
                    entity=entity,
                    reason=reason_string), screen, stacktrace)

            time.sleep(polling)


def satisfied(entity, condition):
    try:
        value = condition(entity)
        return value if value is not None else False
    except Exception as exc:
        return False



import contextlib
import os

from selenium.webdriver.common.by import By


@contextlib.contextmanager
def suppress(*exceptions):
    try:
        yield
    except exceptions:
        pass


def merge(*dict_args):
    """
    Given any number of dicts, shallow copy and merge into a new dict,
    precedence goes to key value pairs in latter dicts.
    """
    result = {}
    for dictionary in dict_args:
        result.update(dictionary)
    return result


def extend(obj, cls, *init_args, **init_kwargs):
    obj.__class__ = type(obj.__class__.__name__, (obj.__class__, cls), {})
    cls.__init__(obj, *init_args, **init_kwargs)


# todo: think on moving to tools
def take_screenshot(driver, name, save_location='./'):
    """ saves screenshot of the current page via driver, with name, to the save_location """
    # Make sure the path exists.
    path = os.path.abspath(save_location)
    if not os.path.exists(path):
        os.makedirs(path)
    full_path = '%s/%s' % (path, name)
    driver.get_screenshot_as_file(full_path)
    return full_path

    # todo: sometimes screenshooting fails at httplib with CannotSendRequest... consider handling this somehow...
    # todo: and of course find the reason - why... it may depend on browser version...


def css_or_by_to_by(css_selector_or_by):
    # todo: will it work `if isinstance(css_selector_or_by, Tuple[str, str]):` ?
    if isinstance(css_selector_or_by, tuple):
        return css_selector_or_by
    if isinstance(css_selector_or_by, str):
        return (By.CSS_SELECTOR, css_selector_or_by)
    raise TypeError('css_selector_or_by should be str with CSS selector or Tuple[by:str, value:str]')



from selene import bys

css = bys.by_css
xpath = bys.by_xpath
name = bys.by_name
link_text = bys.by_link_text
partial_link_text = bys.by_partial_link_text

text = bys.by_text
partial_text = bys.with_text


be_following_sibling = bys.following_sibling
be_parent = bys.parent
be_first_child = bys.first_child



from abc import ABCMeta, abstractmethod, abstractproperty

from future.utils import with_metaclass


class ISeleneWebElementLocator(with_metaclass(ABCMeta, object)):

    @abstractmethod
    def find(self):
        # type: () -> WebElement
        pass

    @abstractproperty
    def description(self):
        # type: () -> str
        pass

    def __str__(self):
        return self.description


class ISeleneListWebElementLocator(with_metaclass(ABCMeta, object)):

    @abstractmethod
    def find(self):
        # type: () -> List[WebElement]
        pass

    @abstractproperty
    def description(self):
        # type: () -> str
        pass

    def __str__(self):
        return self.description





from selenium.webdriver.common.by import By

__author__ = 'yashaka'


def by(css_selector):
    return by_css(css_selector)


def by_css(css_selector):
    return (By.CSS_SELECTOR, css_selector)


def by_name(name):
    return (By.NAME, name)


def by_link_text(text):
    return (By.LINK_TEXT, text)


def by_partial_link_text(text):
    return (By.PARTIAL_LINK_TEXT, text)


def by_xpath(xpath):
    return (By.XPATH, xpath)


def following_sibling():
    return by_xpath("./following-sibling::*")


def parent():
    return by_xpath("..")


def first_child():
    return by_xpath("./*[1]")


def by_text(element_text):
    return by_xpath('.//*[text()[normalize-space(.) = '
                    + escape_text_quotes_for_xpath(element_text)
                    + ']]')


def by_partial_text(element_text):
    return by_xpath('.//*[text()[contains(normalize-space(.), '
                    + escape_text_quotes_for_xpath(element_text)
                    + ')]]')


with_text = by_partial_text


def escape_text_quotes_for_xpath(text):
    return 'concat("", "%s")' % (
        str(
            "\", '\"', \"".join(
                text.split('"'))))



from abc import abstractproperty, abstractmethod

from selenium.webdriver.common.by import By
from selenium.webdriver.remote.webelement import WebElement

from selene.abctypes.search_context import ISearchContext


class IWebElement(ISearchContext):
    @abstractproperty
    def tag_name(self): pass

    @abstractproperty
    def text(self): pass

    @abstractmethod
    def click(self): pass

    @abstractmethod
    def submit(self): pass

    @abstractmethod
    def clear(self): pass

    @abstractmethod
    def get_attribute(self, name): pass

    @abstractmethod
    def is_selected(self): pass

    @abstractmethod
    def is_enabled(self): pass

    @abstractmethod
    def find_element_by_id(self, id_): pass

    @abstractmethod
    def find_elements_by_id(self, id_): pass

    @abstractmethod
    def find_element_by_name(self, name): pass

    @abstractmethod
    def find_elements_by_name(self, name): pass

    @abstractmethod
    def find_element_by_link_text(self, link_text): pass

    @abstractmethod
    def find_elements_by_link_text(self, link_text): pass

    @abstractmethod
    def find_element_by_partial_link_text(self, link_text): pass

    @abstractmethod
    def find_elements_by_partial_link_text(self, link_text): pass

    @abstractmethod
    def find_element_by_tag_name(self, name): pass

    @abstractmethod
    def find_elements_by_tag_name(self, name): pass

    @abstractmethod
    def find_element_by_xpath(self, xpath): pass

    @abstractmethod
    def find_elements_by_xpath(self, xpath): pass

    @abstractmethod
    def find_element_by_class_name(self, name): pass

    @abstractmethod
    def find_elements_by_class_name(self, name): pass

    @abstractmethod
    def find_element_by_css_selector(self, css_selector): pass

    @abstractmethod
    def find_elements_by_css_selector(self, css_selector): pass

    @abstractmethod
    def send_keys(self, *value): pass

    # RenderedWebElement Items
    @abstractmethod
    def is_displayed(self): pass

    @abstractproperty
    def location_once_scrolled_into_view(self): pass

    @abstractproperty
    def size(self): pass

    def value_of_css_property(self, property_name): pass

    @abstractproperty
    def location(self): pass

    @abstractproperty
    def rect(self): pass

    @abstractproperty
    def screenshot_as_base64(self): pass

    @abstractproperty
    def screenshot_as_png(self): pass

    @abstractmethod
    def screenshot(self, filename): pass

    @abstractproperty
    def parent(self): pass

    @abstractproperty
    def id(self): pass

    @abstractmethod
    def __eq__(self, element): pass

    @abstractmethod
    def __ne__(self, element): pass

    @abstractmethod
    def find_element(self, by=By.ID, value=None): pass

    @abstractmethod
    def find_elements(self, by=By.ID, value=None): pass

    @abstractmethod
    def __hash__(self): pass


IWebElement.register(WebElement)




from setuptools import setup
from selene import version

description = 'Concise API for selenium in Python + Ajax support + PageObjects (Selenide port from Java to Python)'
# long_description = 'see http://github.com/yashaka/selene/ for more docs...'

setup(
    name='selene',
    version=version.VERSION,
    url='http://github.com/yashaka/selene/',
    download_url='https://github.com/yashaka/selene/tarball/' + version.VERSION,
    license='MIT',
    author='Iakiv Kramarenko',
    author_email='yashaka@gmail.com',
    description=description,
    # long_description=long_description,
    packages=['selene',
              'selene.abctypes',
              'selene.common',
              'selene.support',
              'selene.support.conditions'],
    include_package_data=True,
    install_requires=['selenium==2.53.1', 'webdriver_manager', 'future', 'backports.functools_lru_cache'],
    platforms='any',
    zip_safe=False,
    keywords=['testing', 'selenium', 'selenide', 'browser', 'pageobject', 'widget', 'wrapper'],
    classifiers=[
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
        'Development Status :: 3 - Alpha',
        'Natural Language :: English',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
        'Topic :: Internet :: WWW/HTTP :: Browsers',
        'Topic :: Software Development :: Testing'
    ]
)



from abc import ABCMeta


def _make_delegator_method(name):
    def delegator(self, *args, **kwargs):
        return getattr(self.__delegate__, name)(*args, **kwargs)
        # todo: consider using __call__() instead of __delegate__
        # in Python delegates are objects with __call__ method..
        # so why not to use the following:
        #     return getattr(self(), name)(*args, **kwargs)
        # ?
    return delegator


# def _make_delegator_method_to_property(name):
#     def delegator(self, *args, **kwargs):
#         return getattr(self.__delegate__, name)
#     return delegator


def _make_delegator_property(name):
    return property(lambda self: getattr(self.__delegate__, name))


def _is_property(name, cls):
    return isinstance(getattr(cls, name, None), property)


class DelegatingMeta(ABCMeta):
    def __new__(mcs, name, bases, dct):
        abstract_property_names = frozenset.union(*(frozenset(filter(lambda m: _is_property(m, base), base.__abstractmethods__))
                                                    for base in bases))

        for base in bases:
            base.__abstractmethods__ = frozenset(filter(lambda m: not _is_property(m, base), base.__abstractmethods__))

        abstract_method_names = frozenset.union(*(base.__abstractmethods__
                                                  for base in bases))

        for name in abstract_method_names:
            if name not in dct:
                dct[name] = _make_delegator_method(name)

        # for name in abstract_property_names:
        #     if name not in dct:
        #         dct[name] = _make_delegator_method_to_property(name)

        cls = super(DelegatingMeta, mcs).__new__(mcs, name, bases, dct)

        for name in abstract_property_names:
            if name not in dct:
                setattr(cls, name, _make_delegator_property(name))

        return cls


# todo: finalize naming: Delegating, Delegate, actual_delegate, delegatee, delegator o_O ?
# We have the following players in this game:
# * MetaClass for Classes of Objects who delegates their implementation to aggregated object
# So who should be named how?



from _ast import Dict, List
from _ctypes import Union
from abc import abstractmethod, abstractproperty
from contextlib import contextmanager

from selenium.webdriver.common.alert import Alert
from selenium.webdriver.common.html5.application_cache import ApplicationCache
from selenium.webdriver.remote.file_detector import FileDetector
from selenium.webdriver.remote.mobile import Mobile
from selenium.webdriver.remote.switch_to import SwitchTo
from selenium.webdriver.remote.webdriver import WebDriver
from selenium.webdriver.remote.webelement import WebElement

from selene.abctypes.search_context import ISearchContext


class IWebDriver(ISearchContext):
    @abstractmethod
    def __repr__(self):
        # type: () -> str
        pass

    @contextmanager
    @abstractmethod
    def file_detector_context(self, file_detector_class, *args, **kwargs):
        pass

    @abstractproperty
    def mobile(self):
        # type: () -> Mobile
        pass

    @abstractproperty
    def name(self):
        # type: () -> str
        pass

    @abstractmethod
    def start_client(self):
        # type: () -> None
        pass

    @abstractmethod
    def stop_client(self):
        # type: () -> None
        pass

    @abstractmethod
    def start_session(self, desired_capabilities, browser_profile=None):
        # type: () -> None
        pass

    @abstractmethod
    def _wrap_value(self, value):
        pass

    @abstractmethod
    def create_web_element(self, element_id):
        # type: (object) -> WebElement
        # todo: object? str?
        pass

    @abstractmethod
    def _unwrap_value(self, value):
        pass

    @abstractmethod
    def execute(self, driver_command, params=None):
        # type: (str, Dict) -> Dict
        pass

    @abstractmethod
    def get(self, url):
        # type: (str) -> Dict
        pass

    @abstractproperty
    def title(self):
        # type: () -> str
        pass

    @abstractmethod
    def find_element_by_id(self, id_):
        # type: (str) -> WebElement
        pass

    @abstractmethod
    def find_elements_by_id(self, id_):
        # type: (str) -> List[WebElement]
        pass

    @abstractmethod
    def find_element_by_xpath(self, xpath):
        # type: (str) -> WebElement
        pass

    @abstractmethod
    def find_elements_by_xpath(self, xpath):
        # type: (str) -> List[WebElement]
        pass

    @abstractmethod
    def find_element_by_link_text(self, link_text):
        # type: (str) -> WebElement
        pass

    @abstractmethod
    def find_elements_by_link_text(self, text):
        # type: (str) -> List[WebElement]
        pass

    @abstractmethod
    def find_element_by_partial_link_text(self, link_text):
        # type: (str) -> WebElement
        pass

    @abstractmethod
    def find_elements_by_partial_link_text(self, link_text):
        # type: (str) -> List[WebElement]
        pass

    @abstractmethod
    def find_element_by_name(self, name):
        # type: (str) -> WebElement
        pass

    @abstractmethod
    def find_elements_by_name(self, name):
        # type: (str) -> List[WebElement]
        pass

    @abstractmethod
    def find_element_by_tag_name(self, name):
        # type: (str) -> WebElement
        pass

    @abstractmethod
    def find_elements_by_tag_name(self, name):
        # type: (str) -> List[WebElement]
        pass

    @abstractmethod
    def find_element_by_class_name(self, name):
        # type: (str) -> WebElement
        pass

    @abstractmethod
    def find_elements_by_class_name(self, name):
        # type: (str) -> List[WebElement]
        pass

    @abstractmethod
    def find_element_by_css_selector(self, css_selector):
        # type: (str) -> WebElement
        pass

    @abstractmethod
    def find_elements_by_css_selector(self, css_selector):
        # type: (str) -> List[WebElement]
        pass

    @abstractmethod
    def execute_script(self, script, *args):
        # type: (str, *object) -> str
        pass

    @abstractmethod
    def execute_async_script(self, script, *args):
        # type: (str, *object) -> str
        pass

    @abstractproperty
    def current_url(self):
        # type: () -> str
        pass

    @abstractproperty
    def page_source(self):
        # type: () -> str
        pass

    @abstractmethod
    def close(self):
        # type: () -> None
        pass

    @abstractmethod
    def quit(self):
        # type: () -> None
        pass

    @abstractproperty
    def current_window_handle(self):
        # type: () -> str
        pass

    @abstractproperty
    def window_handles(self):
        # todo: add type annotation
        pass

    @abstractmethod
    def maximize_window(self):
        # type: () -> None
        pass

    @abstractproperty
    def switch_to(self):
        # type: () -> SwitchTo
        pass

    # *** Target Locators ***

    @abstractmethod
    def switch_to_active_element(self):
        # type: () -> object
        # todo: object? WebElement?
        pass

    @abstractmethod
    def switch_to_window(self, window_name):
        # type: () -> None
        pass

    @abstractmethod
    def switch_to_frame(self, frame_reference):
        # type: () -> None
        pass

    @abstractmethod
    def switch_to_default_content(self):
        # type: () -> None
        pass

    @abstractmethod
    def switch_to_alert(self):
        # type: () -> Alert
        pass

    # *** Navigation ***

    @abstractmethod
    def back(self):
        # type: () -> None
        pass

    @abstractmethod
    def forward(self):
        # type: () -> None
        pass

    @abstractmethod
    def refresh(self):
        # type: () -> None
        pass

    # *** Options ***

    @abstractmethod
    def get_cookies(self):
        # type: () -> Union[str, unicode]
        # todo: Union? Dict?
        pass

    @abstractmethod
    def get_cookie(self, name):
        # todo: type?
        pass

    @abstractmethod
    def delete_cookie(self, name):
        # type: () -> None
        pass

    @abstractmethod
    def delete_all_cookies(self):
        # type: () -> None
        pass

    @abstractmethod
    def add_cookie(self, cookie_dict):
        # type: () -> None
        pass

    # *** Timeouts ***

    @abstractmethod
    def implicitly_wait(self, time_to_wait):
        # type: (int) -> None
        pass

    @abstractmethod
    def set_script_timeout(self, time_to_wait):
        # type: (int) -> None
        pass

    @abstractmethod
    def set_page_load_timeout(self, time_to_wait):
        # type: (int) -> None
        pass

    # @abstractmethod
    # def find_element(self, by=By.ID, value=None):
    #     pass
    #
    # @abstractmethod
    # def find_elements(self, by=By.ID, value=None):
    #     pass

    @abstractproperty
    def desired_capabilities(self):
        # type: () -> Dict
        pass

    @abstractmethod
    def get_screenshot_as_file(self, filename):
        # type: (str) -> bool
        pass

    save_screenshot = get_screenshot_as_file

    def get_screenshot_as_png(self):
        # type: () -> object
        # todo: type?
        pass

    @abstractmethod
    def get_screenshot_as_base64(self):
        # todo: type?
        pass

    @abstractmethod
    def set_window_size(self, width, height, windowHandle='current'):
        # type: (object, object, str) -> None
        pass

    @abstractmethod
    def get_window_size(self, windowHandle='current'):
        # type: (str) -> Dict
        pass

    @abstractmethod
    def set_window_position(self, x, y, windowHandle='current'):
        # type: (object, object, str) -> None
        pass

    @abstractmethod
    def get_window_position(self, windowHandle='current'):
        # type: (str) -> Dict
        pass

    @abstractproperty
    def file_detector(self):
        # type: () -> FileDetector
        pass

    @file_detector.setter
    @abstractmethod
    def file_detector(self, detector):
        # type: (FileDetector) -> None
        pass

    @abstractproperty
    def orientation(self):
        # todo: type?
        pass

    @orientation.setter
    @abstractmethod
    def orientation(self, value):
        # todo: type?
        pass

    @abstractproperty
    def application_cache(self):
        # type: () -> ApplicationCache
        pass

    @abstractproperty
    def log_types(self):
        # todo: type?
        pass

    @abstractmethod
    def get_log(self, log_type):
        # todo: type?
        pass

IWebDriver.register(WebDriver)
